#!/usr/bin/env python

"""
Fix name-related operations. The code in this module operates upon nodes which
are produced when simplifying AST node trees originating from the compiler
module.

Copyright (C) 2006 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from simplified import *
import compiler

# Fixing of name-related operations.

class Fixer(Visitor):

    """
    The name fixer which traverses the program nodes, typically depth-first,
    and maintains a record of name usage in the different namespaces. As a
    consequence of various observations, some parts of the program node tree are
    modified with different operations employed to those originally defined.
    """

    def __init__(self):
        Visitor.__init__(self)

        # Satisfy visitor issues.

        self.visitor = self

    def process_all(self, visitor):
        subprograms = []
        for subprogram in visitor.subprograms:
            subprograms.append(self.process(subprogram))
        visitor.subprograms = subprograms
        visitor.result = self.process(visitor.result)
        return visitor

    def process(self, node):

        """
        Process a subprogram or module 'node', indicating any initial 'locals'
        and 'globals' if either are defined. Return an annotated subprogram or
        module. Note that this method may mutate nodes in the original program.
        """

        # Obtain a namespace either based on locals or on a structure.

        self.namespace = Namespace(structure=getattr(node, "structure", None))

        # Add namespace details to any structure involved.

        if hasattr(node, "structure") and node.structure is not None:

            # Initialise bases where appropriate.

            if hasattr(node.structure, "bases"):
                bases = []
                for base in node.structure.bases:
                    bases.append(self.dispatch(base))
                node.structure.bases = bases

        # Dispatch to the code itself.

        result = self.dispatch(node)
        return result

    # Visitor methods.

    def default(self, node):

        """
        Process the given 'node', given that it does not have a specific
        handler.
        """

        for attr in ("args",):
            value = getattr(node, attr, None)
            if value is not None:
                setattr(node, attr, self.dispatches(value))
        for attr in ("expr", "lvalue", "test", "handler", "star", "dstar"):
            value = getattr(node, attr, None)
            if value is not None:
                setattr(node, attr, self.dispatch(value))
        for attr in ("body", "else_", "finally_", "code"):
            value = getattr(node, attr, None)
            if value is not None:
                setattr(node, attr, self.dispatches(value))
        return node

    def dispatch(self, node, *args):
        return Visitor.dispatch(self, node, *args)

    def visitGlobal(self, global_):
        for name in global_.names:
            self.namespace.make_global(name)
        return global_

    def visitLoadName(self, loadname):
        scope = self.namespace.find_for_load(loadname.name)
        if scope == "structure":
            result = self.dispatch(LoadAttr(expr=LoadRef(ref=self.namespace.structure), name=loadname.name))
        elif scope == "global":
            result = self.dispatch(LoadGlobal(name=loadname.name))
        else:
            result = loadname
        return result

    def visitStoreName(self, storename):
        scope = self.namespace.find_for_store(storename.name)
        if scope == "structure":
            return self.dispatch(StoreAttr(lvalue=LoadRef(ref=self.namespace.structure), name=storename.name, expr=storename.expr))
        elif scope == "global":
            return self.dispatch(StoreGlobal(name=storename.name, expr=storename.expr))
        else:
            storename.expr = self.dispatch(storename.expr)
            self.namespace.store(storename.name)
            return storename

# vim: tabstop=4 expandtab shiftwidth=4
